#-------------------------------------------
#
# Copyright (c) 2001 DataChannel, Inc. All Rights Reserved.
#
# LGPL 2.1
#
# $Id:$
#
#-------------------------------------------

import sys, os, os.path
import string
import traceback
from StringIO import StringIO

import ZODB
import Persistence

class StorageList(Persistence.Persistent):
    """Just a Persistent object to hold a list of StorageObjects"""
    def __init__(self):
        self.__list = {}

    def getStorageObjects(self):
        return self.__list.values()

    def getStorageObject(self, name):
        return self.__list[name]

    def addStorageObject(self, storage):
        self.__list[storage.getName()] = storage
        self.__changed__(1)


class _Block(Persistence.Persistent):
    """Internal Persistent datastructure used to help persist StorageObject
    more efficiently"""
    def __init__(self, data=''):
        self.data = data
        self.__changed__(1)

    def set(self, data):
        self.data = data
        self.__changed__(1)

    def __len__(self):
        return len(self.data)

    def __getslice__(self, a, b):
        return self.data[a:b]
        
    def __delslice__(self, a, b):
        old = self.data
        self.data = self.data[:a] + self.data[b:]
        if self.data != old:
            self.__changed__(1)

class StorageObject(Persistence.Persistent):
    """This class holds the storage data.

    Implements the standard Python I/O methods (as for cStringIO).
    """

    def __init__(self, name='', buf=''):
        self.__name = name

        self.blocks = []
        self.length = 0

        if buf:
            self[:] = buf

    def setName(self, name):
        self.__name = name

    def getName(self):
        return self.__name

    def size(self):
        return self._size()

    def _size(self):
        return len(self)

    def __len__(self):
        return self.length

    def __decLength(self, amount):
        self.length = self.length - amount

    def __incLength(self,amount):
        self.length = self.length + amount
    
    def read(self, offset, size=0):
        """If size == 0 then read till end"""
        return self._read(offset, size)

    def _read(self, offset, size=0):
        if not size:
            return self[offset:]
        else:
            return self[offset:offset+size]
    
    def __getslice__(self, a, b):
        str = StringIO()

        currentBlockIndex, startStringIndex = self._findIndex(a)
        endBlockIndex, endStringIndex = self._findIndex(b)
        
        while (currentBlockIndex <= endBlockIndex):
            if currentBlockIndex == endBlockIndex and \
               startStringIndex == endStringIndex:
                break
            currentBlock = self.blocks[currentBlockIndex]
            if currentBlockIndex == endBlockIndex:
                str.write(currentBlock[startStringIndex:endStringIndex])
            else:
                str.write(currentBlock[startStringIndex:])

            # increment loop vars
            currentBlockIndex = currentBlockIndex + 1
            startStringIndex = 0

        return str.getvalue()

    def write(self, offset, buf):
        return self._write(offset, buf)

    def _write(self, offset, buf):
        self[offset:offset+len(buf)] = buf
        
    def __setslice__(self, a, b, v):
        del self[a:b]

        if not v:
            return
        
        blockIndex, stringIndex = self._findIndex(a)
        
        self.__incLength(len(v))
        self.blocks.insert(blockIndex+1, _Block(v))

    def truncate(self, offset, size=0):
        return self._truncate(offset, size)

    def _truncate(self, offset, size=0):
        if not size:
            del self[offset:]
        else:
            del self[offset:offset+size]

    def __delslice__(self, a, b):
        startBlockIndex, startStringIndex = self._findIndex(a)
        endBlockIndex, endStringIndex = self._findIndex(b)

        # check range
        if len(self.blocks) <= startBlockIndex:
            return

        # handle start block
        startBlock = self.blocks[startBlockIndex]
        startBlockLen = len(startBlock)
        if startStringIndex < startBlockLen:
            self.__decLength( startBlockLen - startStringIndex )
            del startBlock[startStringIndex:]
        if startStringIndex > 0:
            startBlockIndex = startBlockIndex + 1

        # handle end block
        endBlock = self.blocks[endBlockIndex]
        endBlockLen = len(endBlock)
        if endStringIndex < endBlockLen:
            self.__decLength( len(endBlock[:endStringIndex]) )
            del endBlock[:endStringIndex]
        else:
            endBlockIndex = endBlockIndex + 1

        # handle self.blocks
        for block in self.blocks[startBlockIndex:endBlockIndex]:
            self.__decLength(len(block))
        del self.blocks[startBlockIndex:endBlockIndex]

    def _findIndex(self, index):
        "_findIndex(self, index) -> (blockIndex, stringIndex)"
        currentBlockIndex = 0
        currentStringIndex = 0

        if index == 0:
            return (currentBlockIndex, currentStringIndex)

        while currentStringIndex <= index \
              and len(self.blocks) > currentBlockIndex:
            currentBlock = self.blocks[currentBlockIndex]

            nextCurrentStringIndex = currentStringIndex + len(currentBlock)
            
            if index <= nextCurrentStringIndex:
                return (currentBlockIndex, index-currentStringIndex)

            # increment loop vars
            currentStringIndex = nextCurrentStringIndex
            currentBlockIndex = currentBlockIndex + 1


        currentBlockIndex = currentBlockIndex - 1 # cause last loop
        return (currentBlockIndex, currentStringIndex)
