#-------------------------------------------
#
# Copyright (c) 2001 DataChannel, Inc. All Rights Reserved.
#
# LGPL 2.1
#
# $Id:$
#
#-------------------------------------------

import sys
import thread
import time
import random
import traceback
import tprofile
from StringIO import StringIO

import omniORB
from omniORB import CORBA, PortableServer, PortableServer__POA


from TransactionManagement import Transaction

import ZODB
from BTrees.OOBTree import OOBTree
from Persistence import Persistent
from ZODB.POSException import ConflictError

from zLOG import LOG, TRACE, DEBUG, BLATHER, INFO, PROBLEM, ERROR, PANIC

def formatException(exception):
    """
    Used to return a nice stack trace with our corba exceptions
    """
    strIO = StringIO()
    traceback.print_exc(file=strIO)
    return strIO.getvalue()

MAX_COMMIT_ATTEMPTS=10

## Setup Domain to Servant Map
# This is a map from Domain class to Servant class
# Every servant should have an entry here.

_domainToServant = None
def getDomainToServantMap():
    """Returns a mapping of domain class to servant class."""
    global _domainToServant
    if not _domainToServant:
        _domainToServant = {}

    return _domainToServant

_corbaTypeIdToDomain = None
_corbaDomainToTypeId = None

def getCorbaTypeIdToDomainMap():
    """
    Returns a mapping of servant class name to domain class.
    """
    global _corbaTypeIdToDomain
    if not _corbaTypeIdToDomain:
        _corbaTypeIdToDomain = {}
    return _corbaTypeIdToDomain

def getDomainToCorbaTypeIdMap():
    """
    """
    global _corbaDomainToTypeId
    if not _corbaDomainToTypeId:
        _corbaDomainToTypeId = {}
    return _corbaDomainToTypeId



def registerDomainToServantMap(domainKlass, servantKlass):
    """
    Registers a mapping from the given domain class to servant class.
    Registers a mapping from repository id of the servant class
    to the domain class.
    """
    getDomainToServantMap()[domainKlass] = servantKlass
    getCorbaTypeIdToDomainMap()[CORBA.id(servantKlass)] = domainKlass
    getDomainToCorbaTypeIdMap()[domainKlass] = CORBA.id(servantKlass)

def getDomainForCorbaTypeId(typeId):
    """
    This function returns a Domain Class to be used for the
    given corba repository Id.
    If no Domain Class has been registered.
    """
    return getCorbaTypeIdToDomainMap().get(typeId, None)

def getCorbaTypeIdForDomain(domainKlass):
    return getDomainToCorbaTypeIdMap().get(domainKlass, None)

def getServantForDomain(domainKlass):
    """
    This function returns a Servant Class to be used for the given domain Class.
    If no Servant Class has been registered, The domain Class's base classes
    are searched for a corresponding servant.
    """

    servantKlass = None
    servantKlass = getDomainToServantMap().get(domainKlass, None)
    #FIXME should use stack, should probably give precedence to
    #immediate superclasses
    if not servantKlass:
        bases = domainKlass.__bases__
        for base in bases:
            servantKlass = getServantForDomain(base)
            if servantKlass:
                break
    return servantKlass


## End Setup Domain to Servant Map

### Setup Python exception to CORBA exception factory map
#    The exception factory is a function which produces CorbaExceptions for a
#    given python exception.  This will only be called if the
#    exceptionFactory has been registered in the global
#    pythonToCORBAExceptionFactoryMap with the pythonException as a key.
###

_defaultCORBAExcFactory = None
_pythonToCORBAExcFactory = None
def getPythonToCORBAExcFactoryMap():
    """Returns a mapping of python exception to CORBA exception factory."""
    global _pythonToCORBAExcFactory
    if not _pythonToCORBAExcFactory:
        _pythonToCORBAExcFactory = {}

    return _pythonToCORBAExcFactory


def getFactoryForPyExc(domainExceptionKlass):
    """
    This function returns a Corba Exception Factory (CEF) to be used
    for the given domain Exception Class.
    The Corba Exception Factory takes an exception as an argument and
    processes it by throwing an appropriate Corba Exception.
    If no domain Exception Class has been registered, The domain
    Exception Class's base classes are searched for a corresponding CEF.
    """
    factory = None
    factory = getPythonToCORBAExcFactoryMap().get(
        domainExceptionKlass,
        None)
    #FIXME should use stack, should probably give precedence to
    #immediate superclasses
    if not factory:
        bases = domainExceptionKlass.__bases__
        for base in bases:
            factory = getFactoryForPyExc(base)
            if factory:
                break

    if not factory and _defaultCORBAExcFactory:
        factory = _defaultCORBAExcFactory
    return factory

def registerDefaultCORBAExcFactory(CORBAExcFactory):
    global _defaultCORBAExcFactory
    _defaultCORBAExcFactory = CORBAExcFactory

def registerPythonToCORBAExcFactoryMap(pythonExc, CORBAExcFactory):
    """Registers a mapping of the given python exception to the
    given CORBA exception factory."""
    getPythonToCORBAExcFactoryMap()[pythonExc] = CORBAExcFactory

def _getDomainId(domainObj):
    """Returns the domain id for the given domain object."""
    if hasattr(domainObj, 'hash'):
        return str(hash(domainObj))
    else:
        return str(id(domainObj))

class ServantHelper:
    """Documentation interface for Servants constructor
    requirement.
    Implemented in the ServantLocator"""
    def getObjRef(self, obj):
        pass
    def getObjRefList(self, objList):
        pass
    def objRefGenerator(self, id, obj, servantKlass):
        pass
    def lookupDomainObj(self, objRef):
        pass
    def removeDomainObj(self, domainObj):
        pass

class BaseServant:
    def _raiseCorbaException(self, exc):
        if isinstance(exc, CORBA.Exception):
            raise

        factory = getFactoryForPyExc(exc.__class__)
        if factory:
            try:
                factory(exc, self)
            except:
                LOG("ZODBCorbaFramework",
                    INFO,
                    "Raising CORBA Exception",
                    error=sys.exc_info())
                raise
        else:
            raise

    def _tryFunction(self, func, args):
        """
        Applys the given function to the arguments and returns
        the result.
        Throws an appropriate exception upon failure.
        """
        try:
            result = apply(func, args)
            return result
        except ConflictError:
            raise
        except Exception, exc:
            self._raiseCorbaException(exc)

    def _tryCommitFunction(self, func, args):
        """
        Uses _tryFunction to apply the given function to the
        arguments and then performs a commit.
        Throws an appropriate exception upon failure.
        """
        sleepTime=.05
        for i in range(MAX_COMMIT_ATTEMPTS):
            try:
                result = self._tryFunction(func, args)
                get_transaction().commit()
                return result

            except ConflictError:
                get_transaction().abort()
                time.sleep(sleepTime)
                sleepTime = sleepTime * 1.5
            except:
                raise

        exc = ConflictError("Couldn't commit in %s tries!" % \
                            MAX_COMMIT_ATTEMPTS)
        self._raiseCorbaException(exc)

class _Functor:
    def __init__(self, servant):
        self.servant = servant
    def __call__(self, *args, **kwArgs):
        try:
            callable = getattr(self.servant._domainObj,
                               self.servant._methodName)
        except AttributeError, exc:
            raise AttributeError(self.servant._domainObj,
                                 self.servant._methodName,
                                 *exc.args)
        return apply(callable, args, kwArgs)

class Servant(BaseServant):
    """Abstract Servant that takes a domain object to wrap
    and a ServantHelper to get object reference things done."""
    def __init__(self, servantHelper):
        self._servantHelper = servantHelper
        self._domainObj = None
        self.__functor = _Functor(self)
        self._method = None

    def _load(self, domainObj):
        self._domainObj = domainObj

    def _empty(self):
        self._domainObj = None

    def _tryDomainCommit(self, methodName, *args):
        self._methodName = methodName
        try:
            return self._tryCommitFunction(self.__functor, *args)
        finally:
            self._methodName = None

class RemoveableServant(Servant):
    def remove(self):
        self._servantHelper.removeDomainObj(self._domainObj)

_SESSION_DATA_MAP = '_session_data_map'

class _List(Persistent):
    def __init__(self):
        self.__version = "1.0"
        self.list = []
        self.next = None
    def append(self, obj):
        self.list.append(obj)
        self.__changed__(1)
    def __len__(self):
        return len(self.list)

class SessionData(Persistent):
    def __init__(self):
        self.__version = "1.0"
        self.listStart = _List()
        self.listEnd = self.listStart
        
    def addObject(self, obj):
        if len(self.listEnd) > 10:
            old = self.listEnd
            self.listEnd = _List()
            old.next = self.listEnd
            
        self.listEnd.append(obj)

class ZODBCorbaServer(BaseServant):
    """
    Subclasses must also sublcass a CORBA servant.
    """

    def __init__(self, storage, serverName=None):
        # Initialise the ORB and activate the root POA.
        orb = CORBA.ORB_init(sys.argv, CORBA.ORB_ID)
        rootPOA = orb.resolve_initial_references("RootPOA")
        rootPOAManager = rootPOA._get_the_POAManager()
        rootPOAManager.activate()

        if not serverName:
            tmp = str(time.time())
            serverName = "ZODBCorbaServer" + tmp

        self._poa = orb.resolve_initial_references("omniINSPOA")
        self._poaManager = self._poa._get_the_POAManager()
        self._poaManager.activate()

        self._poa.activate_object_with_id(serverName, self)

        self.__initDB(ZODB.DB(storage))

    def __initDB(self, db):
        self._transaction = Transaction()
        self._transaction.attach()
        self._db = db

        conn = db.open()
        root = conn.root()
        try:
            self._tryCommitFunction(self.__initDBData,
                                (root,))
            self._initDB(root)
        finally:
            conn.close()

        self._transaction.detach()

        self._db.setPoolSize(10000)
        self._db.setCacheSize(10000)


    def __initDBData(self, root):
        if not root.has_key(_SESSION_DATA_MAP):
            root[_SESSION_DATA_MAP] = OOBTree()

    def _initDB(self, root):
        pass

    def _getDB(self):
        return self._db

    def _default_POA(self):
        return self._poa

    def getSession(self, *args):
        return self._tryFunction(
            self._getSession, args).getObjRef()

    def _getSession(self, args):
        return Session(self._db)

class Session(BaseServant):
    """
    Subclasses must also sublcass a CORBA servant.  Mimics
    interface of Servant class to use servantLocator
    """

    def __init__(self, db, *args):
        self.__rootPOA = CORBA.ORB_init().resolve_initial_references("RootPOA")
        self.__initDB(db, *args)
        self.__initPOA(*args)

    def __initDB(self, db, *args):
        self._trans = Transaction()
        self._connection = db.open()
        self._root = self._connection.root()

        self._callbackPreInvoke()
        self._tryCommitFunction(self.__initDBData, ())

        try:
            self._initDB(db, *args)
        finally:
            self._callbackPostInvoke()

    def __initDBData(self):
        sessionDataMap = self._root[_SESSION_DATA_MAP]
        self._sessionDataKey = random.random()
        if sessionDataMap.has_key(self._sessionDataKey):
            raise ConflictError()
        self._sessionData = SessionData()
        sessionDataMap[self._sessionDataKey] = self._sessionData

    def _initDB(self, db, *args):
        pass

    def __initPOA(self, *args):
        self.__objRef = None
        self._poa, self._servantLocator = self._createPoaAndServantLocator()

        # set Session Servant reference
        self._servantLocator.setSession(self)

        self._callbackPreInvoke()
        try:
            self._initPOA(*args)
        finally:
            self._callbackPostInvoke()

    def _initPOA(*args):
        pass

    def _default_POA(self):
        return self._poa

    def getObjRef(self):
        if not self.__objRef:
            oid = str(id(self))
            self._servantLocator.addFixedServant(oid, self)
            corbaId = CORBA.id(self.__class__)
            self.__objRef = self._poa.create_reference_with_id(
                oid,
                corbaId)
        return self.__objRef

    def _getDBConnection(self):
        if not self._connection:
            self._connection = None
        return self._connection

    def _callbackPreInvoke(self):
        """This will setup global thread indexes to transaction
        for the current Thread so that all Persistent object
        modifications will be registerd on the right Transaction."""
        try:
            self._trans.attach()

            conn = self._getDBConnection()
            conn.sync()
        except:
            exc = sys.exc_info()
            try:
                self._callbackPostInvoke()
            except:
                pass
            raise

    def _callbackPostInvoke(self):
        """Called when finished CORBA Request to unregister current
        thread with Transaction"""
        self._trans.detach()

    def logout(self):
        self._tryCommitFunction(self.__removeSessionData, ())
        self._connection.close()
        self._servantLocator.logout()
        self._poa.destroy(0, 0)
        self._callbackPostInvoke()

    def __removeSessionData(self):
        del self._root[_SESSION_DATA_MAP][self._sessionDataKey]

    def _createPoaAndServantLocator(self):
        """Create a child POA with the right policies for a
        ServantLocator"""
        ps = [self.__rootPOA.create_id_assignment_policy(
            PortableServer.USER_ID),
              self.__rootPOA.create_servant_retention_policy(
            PortableServer.NON_RETAIN),
              self.__rootPOA.create_thread_policy(
            PortableServer.SINGLE_THREAD_MODEL),
              self.__rootPOA.create_request_processing_policy(
            PortableServer.USE_SERVANT_MANAGER)]

        tmp = str(time.time()) + str(random.random())

        child = self.__rootPOA.create_POA(tmp, None, ps)

        # Create the ServantLocator and set it as the child's ServantManager
        sli = ServantLocator(child)
        slo = sli._this()
        child.set_servant_manager(slo)
        child._get_the_POAManager().activate()
        return child, sli

class ServantLocator (PortableServer__POA.ServantLocator):
    def __init__(self, poa):
        self.__session = None
        self.__poa = poa
        self.__domainClassToServant = {}
        self.__idToDomainObj = {}
        self.__idToObjRef = {}
        self.__idToServant = {}

    def preinvoke(self, oid, poa, operation):
        #tprofile.profileThread(operation)
        try:
            if not self.__session:
                raise CORBA.OBJECT_NOT_EXIST("Session Does Not exist")
            result = None

            self.__session._callbackPreInvoke()

            if self.__idToServant.has_key(oid):
                servant = self.__idToServant[oid]
                
            elif self.__idToDomainObj.has_key(oid):
                domainObj = self.__idToDomainObj[oid]
                assert not domainObj == None
                servant = self.__getServant(domainObj.__class__)
                if not servant:
                    raise CORBA.OBJECT_NOT_EXIST()
                servant._load(domainObj)
            else:
                raise CORBA.OBJECT_NOT_EXIST()
            return (servant, "")
        except Exception , exc:
            self.__session._raiseCorbaException(exc)
                    
    def postinvoke(self, oid, poa, operation, cookie, serv):
        try:
            try:
                if self.__idToDomainObj.has_key(oid):
                    domainObj = self.__idToDomainObj[oid]
                    servant = self.__domainClassToServant[
                        domainObj.__class__]
                    servant._empty()
            except:
                try:
                    self.__session._raiseCorbaException(exc)
                except:
                    traceback.print_exc()
        finally:
            try:
                if self.__session:
                    self.__session._callbackPostInvoke()
            except:
                traceback.print_exc()
                raise

    def setSession(self, session):
        """
        registers the session object for use by the servantLocator
        """
        self.__session = session

    def addFixedServant(self, oid, servant):
        self.__idToServant[oid] = servant

    def logout(self):
        self.__idToDomainObj.clear()
        self.__idToObjRef.clear()
        self.__idToServant.clear()
        self.__session = None

    def objRefGenerator(self, oid, domainObj, servantClass):
        if not self.__idToObjRef.has_key(oid):
            self.__idToDomainObj[oid] = domainObj
            objRef = self.__poa.create_reference_with_id(
                oid, CORBA.id(servantClass))
            self.__idToObjRef[oid] = objRef

        return self.__idToObjRef[oid]

    def lookupDomainObj(self, objRef):
        if not objRef:
            return None
        try:
            id = self.__poa.reference_to_id(objRef)
        except CORBA.UserException, e:
            #import pdb; pdb.set_trace()
            LOG("ZODBCorbaFramework",
                ERROR,
                """Got CORBA.UserException from reference_to_id
                object reference: %s
                ior: %s""" % \
                (objRef, CORBA.ORB_init().object_to_string(objRef)),
                error=sys.exc_info())
            raise CORBA.OBJ_ADAPTER()

        try:
            return self.__idToDomainObj[id]
        except KeyError:
            LOG("ZODBCorbaFramework",
                ERROR,
                """Couldn't resolve __idToDomainObj[id] where id == %s""" % \
                id,
                error=sys.exc_info())
            raise CORBA.OBJECT_NOT_EXIST()

    def removeDomainObj(self, domainObj):
        corbaId = _getDomainId(domainObj)
        del self.__idToDomainObj[corbaId]
        del self.__idToObjRef[corbaId]

    def __getDomainId(self, domainObj):
        """Returns the domain id for the given domain object."""
        #OPT try using p_oid, else id()
        try:
            return domainObj._p_oid
        except:
            return str(id(domainObj))

    def getObjRef(self, domainObj):
        """Given a domain object, get an id for it and generate
        a CORBA object reference for it."""

        if domainObj is None:
            return None

        oid = _getDomainId(domainObj)

        if not self.__idToObjRef.has_key(oid):
            servantKlass = getServantForDomain(domainObj.__class__)
            objRef = self.__poa.create_reference_with_id(
                oid,
                CORBA.id(servantKlass))
            self.__idToDomainObj[oid] = domainObj
            self.__idToObjRef[oid] = objRef

        return self.__idToObjRef[oid]

    def getObjRefList(self, domainList):
        """Like getObjRef(), but for list"""
        result = []
        for o in domainList:
            result.append(self.getObjRef(o))
        return result

    def lookupDomainObjList(self, objRefList):
        """Utility function to get the domain object from a
        list of object references."""

        domainObjList = []
        if objRefList:
            for objRef in objRefList:
                if objRef != None:
                    domainObjList.append(
                        self.lookupDomainObj(objRef))
        return domainObjList

    def __getServant(self, forDomainKlass):
        result = self.__domainClassToServant.get(forDomainKlass,
                                                 None)

        if not result:
            servantKlass = getServantForDomain(forDomainKlass)
            if servantKlass:
                result = servantKlass(self)
                self.__domainClassToServant[forDomainKlass] = result
        return result
